/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * cddb_query.h
 * A CDDB query interface for looking up discs against a CDDB server. The
 * implementation is defined in cddb_query_impl.h.
 *
 * getResults returns a Discs object, once the query has been executed.
 * There are numerous exceptions that can be thrown by query. If an exception
 * is thrown then the state of the object remains the same.
 */

#ifndef TAGLOOKUP_CDDB_QUERY_H
#define TAGLOOKUP_CDDB_QUERY_H

#include <vector>
#include <exception>

#include "tracks.h"

namespace TagLookup
{
    enum Cache
        {ON, OFF, ONLY};
        
    class CDDBQueryImpl;
    
    typedef std::vector<int> OffsetList;
     
    class CDDBQuery
    {
    public:
        CDDBQuery();
        CDDBQuery(const unsigned int length, const OffsetList &offsets);
        CDDBQuery(const std::string &category, const unsigned int disc_id);

        virtual ~CDDBQuery();
        
        void query(); // queries CDDB, may throw
        
        const Discs & getResults(); 
        
        void setProtocolHTTP(const std::string &hostname, const unsigned int port);
        void setProtocolHTTP(const std::string &hostname, const unsigned int port, const std::string &proxy_hostname, const unsigned int proxy_port, const std::string &proxy_username, const std::string &proxy_password);
        void setProtocolCDDBP(const std::string &hostname, const unsigned int port);
        void setCache(const Cache &cache);
    private:
        // Pimpl
        CDDBQueryImpl *pimpl;
        
    }; // class CDDBQuery
    
    class CDDBQueryException
        : public std::exception
    {
    public:
        CDDBQueryException()
        {
        }
        
        ~CDDBQueryException() throw()
        {
        }

        CDDBQueryException(const std::string &description)
        {
            desc_ = description;
        }
    
        virtual const char * what() const throw()
        {
            return desc_.c_str();
        }
    
    private:
        std::string desc_;
    }; // class CDDBQueryException

}
#endif 



