/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * cddb_query_impl.cpp
 * Implementation of CDDBQueryImpl using libcddb.
 */
 
#include <vector>
#include <iostream>

#include <cddb/cddb.h>

#include "tracks.h"

// Implements:
#include "cddb_query_impl.h"

namespace TagLookup
{
    void CDDBQueryImpl::init()
    {
        /**
         * Initlise the CDDBQueryImpl structure
         */
        
        // Some sensible defaults for the public settable fields
        length_ = 0;
        disc_id_ = 0;

        conn_ = NULL;
        libcddb_init();
        conn_ = cddb_new();
        
        if(conn_ == NULL)
            throw CDDBQueryException("Out of memory when calling cddb_new()");
    }

    CDDBQueryImpl::CDDBQueryImpl()
    {
        init();
    }

    CDDBQueryImpl::CDDBQueryImpl(const unsigned int length, const OffsetList &offsets)
    {
        init();

        length_ = length;
        offsets_ = std::vector<int>(offsets);
    }

    CDDBQueryImpl::CDDBQueryImpl(const std::string &category, const unsigned int disc_id)
    {
        init();

        category_ = category;
        disc_id_ = disc_id;
    }
    
    CDDBQueryImpl::~CDDBQueryImpl()
    {
        /**
         * Destroy the CDDBQueryImpl structure
         */
         
        cddb_destroy(conn_);
        libcddb_shutdown();
    }
    
    void CDDBQueryImpl::query()
    {
        /**
         * Queries CDDB based on the details from:
         * - disc_id and category; or
         * - length and offsets.
         *
         * Can throw various flavours of CDDBQueryException.
         */
        
        cddb_disc_t *disc = NULL;
        
        disc = cddb_disc_new();
        
        if(disc == NULL)
            throw CDDBQueryException("Out of memory when calling cddb_disc_new()");
        
        cddb_track_t *track = NULL;
        int m; // number of matches
        
        if(offsets_.empty()) // use category and disc_id
        {
            cddb_disc_set_category_str(disc, category_.c_str());
            
            if(cddb_disc_get_category(disc) == CDDB_CAT_MISC && category_.compare("misc") != 0)
                std::cerr << "TagLookup: Warning \"" << category_ << "\" matches CDDB_CAT_MISC" << std::endl;
            
            cddb_disc_set_discid(disc, disc_id_);
            m = 1;
        } 
        else // use length and offsets
        {
            cddb_disc_set_length(disc, length_);
            
            OffsetList::const_iterator end = offsets_.end();
            
            for(OffsetList::const_iterator i = offsets_.begin();
                i != end;
                ++i)
            {
                track = NULL;
                track = cddb_track_new();
                if(track == NULL)
                {
                    cddb_track_destroy(track);
                    cddb_disc_destroy(disc);
                    throw(CDDBQueryException("Out of memory whilst calling cddb_track_new()"));
                }
                
                cddb_track_set_frame_offset(track, *i);
                cddb_disc_add_track(disc, track);
            } // end for Offsets
            
            cddb_disc_calc_discid(disc);

            // perform a search
            m = cddb_query(conn_, disc);
            
            if(m == -1)
            {
                // horrible connection error
                std::string msg("Connection error: libcddb: ");
                msg.append(cddb_error_str(cddb_errno(conn_)));
                cddb_disc_destroy(disc);
                
                throw(CDDBQueryException(msg));
            } // if(m == -1)
        } // if query_req
        
        // read each disc and each track from the results
        int success = 0;
        for(int i = 0; i < m; ++i)
        {
            success = cddb_read(conn_, disc);
            
            if(!success)
            {
                // a read or connection error
                std::string msg("Query error: libcddb: ");
                msg.append(cddb_error_str(cddb_errno(conn_)));
                cddb_disc_destroy(disc);
                
                throw(CDDBQueryException(msg));
            } // if(!success)
            
            cddb_query_next(conn_, disc);

            results_.push_back(createDiscFromCDDBDisc(disc));
        } // for each disc in results

        cddb_disc_destroy(disc);
    } // query()
    
    Disc CDDBQueryImpl::createDiscFromCDDBDisc(cddb_disc_t *cd)
    {
        std::string title(cddb_disc_get_title(cd)),
            artist(cddb_disc_get_artist(cd)),
            genre(cddb_disc_get_genre(cd));

        Disc disc(title,
            artist,
            genre,
            cddb_disc_get_year(cd));
        
        cddb_track_t *ct = NULL;

        // Get each track in the disc
        ct = cddb_disc_get_track_first(cd);
        
        while(ct != NULL)
        {
            std::string title(cddb_track_get_title(ct));
            std::string artist(cddb_track_get_artist(ct));
            
            Track track(title,
                artist,
                cddb_track_get_length(ct),
                cddb_track_get_number(ct));
        
            ct = cddb_disc_get_track_next(cd);
            
            disc.appendTrack(track);
        } // while(track != NULL)
        return disc;
    }
    
    void CDDBQueryImpl::setProtocolHTTP(const std::string &hostname, const unsigned int port)
    {
        cddb_http_enable(conn_);
        cddb_set_server_name(conn_, hostname.c_str());
        cddb_set_server_port(conn_, port);
        
        cddb_http_proxy_disable(conn_);
    }
    
    void CDDBQueryImpl::setProtocolHTTP(const std::string &hostname, 
        const unsigned int port, 
        const std::string &proxy_hostname, 
        const unsigned int proxy_port, 
        const std::string &proxy_username, 
        const std::string &proxy_password)
    {
        setProtocolHTTP(hostname, port);
        
        cddb_http_proxy_enable(conn_);
        
        cddb_set_http_proxy_server_name(conn_, proxy_hostname.c_str());

        cddb_set_http_proxy_server_port(conn_, proxy_port);

        cddb_set_http_proxy_credentials(conn_, 
            proxy_username.c_str(), 
            proxy_password.c_str());
    }
    
    void CDDBQueryImpl::setProtocolCDDBP(const std::string &hostname, const unsigned int port)
    {
        cddb_set_server_name(conn_, hostname.c_str());
        cddb_set_server_port(conn_, port);
    }
    
    void CDDBQueryImpl::setCache(const Cache &cache)
    {
        if (cache == ON) {
            cddb_cache_enable(conn_);
        } else if (cache == OFF) {
            cddb_cache_disable(conn_);
        } else if (cache == ONLY) {
            cddb_cache_only(conn_);
        } else {
            throw CDDBQueryException("Unknown cache mode");
        }
    }
    
    const Discs & CDDBQueryImpl::getResults()
    {
        return results_;
    }
}

