/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * command_line.cpp
 * Implementation for command_line.h. This implements a simple recursive-decent 
 * parser. This (crudely) uses each argv as a token.
 */

#include "command_line.h"
#include "command_line_spellings.h"

#include <string>
#include <vector>
#include <algorithm>
#include <utility>

#include <stdlib.h>
#include <limits.h>

namespace TagLookup
{
    void CommandLineArgumentsParser::accept(const std::string &str)
    {
        if(ca_ == argv_.end())
        {
            std::string msg("Expected \"");
            msg.append(str);
            msg.append("\".");
            throw ParseException(msg);
        }

        if(ct_ == str)
        {
            ++ca_;
            if(ca_ == argv_.end())
                ct_ = "";
            else
                ct_ = *ca_;
            return;
        } 
        else
        {
            std::string msg("Expected \"");
            msg.append(str);
            msg.append("\" got ");
            msg.append(ct_);
            throw ParseException(msg);
        } 
    }

    void CommandLineArgumentsParser::acceptEOF()
    {
        if(ca_ == argv_.end())
            return;
        else
            throw ParseException("Expected EOF");
    }
    
    bool CommandLineArgumentsParser::parseArguments()
    {
        accept(ct_);
        parseFlags();
        parseCommand();
        acceptEOF();
        
        return true;
    }
    
    bool CommandLineArgumentsParser::parseFlags()
    {
        while( parseFlag() );
        
        return true;
    }
    
    bool CommandLineArgumentsParser::parseServer()
    {
        if(ct_.empty())
        {
            throw ParseException("Expected server hostname");
        }
        else
        {
            parsed_.server = ct_;
            parsed_.server_specified = true;
            accept(ct_);
        }
        return true;
    }
    
    bool CommandLineArgumentsParser::parsePort()
    {
        int port = (int)std::strtol(ct_.c_str(), (char**)NULL, 10);
        
        if(port > 32767 || port <= 0)
        {
            throw ParseException("Expected valid port number (0-32767)");
        }
        else
        {
            parsed_.port = port;
            parsed_.port_specified = true;
            accept(ct_);
        }
        return true;
    }
    
    bool CommandLineArgumentsParser::parseFlag()
    {
        if(ct_ == TAGLOOKUP_CACHE_MODE_FLAG)
        {
            accept(TAGLOOKUP_CACHE_MODE_FLAG);
            parseCacheMode();
            return true;
        }

        if(ct_ == TAGLOOKUP_SERVER_FLAG)
        {
            accept(TAGLOOKUP_SERVER_FLAG);
            parseServer();
            return true;
        }

        if(ct_ == TAGLOOKUP_PORT_FLAG)
        {
            accept(TAGLOOKUP_PORT_FLAG);
            parsePort();
            return true;
        }
        
        if(ct_ == TAGLOOKUP_PROTOCOL_FLAG)
        {
            accept(TAGLOOKUP_PROTOCOL_FLAG);
            parseProtocol();
            return true;
        }

        return false;
    }
    
    bool CommandLineArgumentsParser::parseCacheMode()
    {
        if(ct_ == TAGLOOKUP_ON)
        {
            parsed_.cache_mode = ON;
            parsed_.cache_mode_specified = true;
            accept(TAGLOOKUP_ON);
            return true;
        } 

        if(ct_ == TAGLOOKUP_OFF)
        {
            parsed_.cache_mode = OFF;
            parsed_.cache_mode_specified = true;
            accept(TAGLOOKUP_OFF);
            return true;
        }

        if(ct_ == TAGLOOKUP_ONLY)
        {
            parsed_.cache_mode = ONLY;
            parsed_.cache_mode_specified = true;
            accept(TAGLOOKUP_ONLY);
            return true;
        }
        
        throw ParseException("Unknown cache mode");
    }

    bool CommandLineArgumentsParser::parseProtocol()
    {
        if(ct_ == TAGLOOKUP_CDDBP)
        {
            parsed_.protocol = TAGLOOKUP_PROTOCOL_CDDBP;
            parsed_.protocol_specified = true;
            accept(TAGLOOKUP_CDDBP);
            return true;
        }     
        if(ct_ == TAGLOOKUP_HTTP)
        {
            parsed_.protocol = TAGLOOKUP_PROTOCOL_HTTP;
            parsed_.protocol_specified = true;
            accept(TAGLOOKUP_HTTP);
            return true;
        }     
        if(ct_ == TAGLOOKUP_PROXY)
        {
            parsed_.protocol = TAGLOOKUP_PROTOCOL_HTTP_PROXY;
            parsed_.protocol_specified = true;
            accept(TAGLOOKUP_PROXY);
            return true;
        }
        
        throw ParseException("Unknown protocol");
    }
    
    bool CommandLineArgumentsParser::parseCDDBCategory()
    {
        if(ct_.empty())
            throw ParseException("Expected CDDB category");
        else
        {
            parsed_.category = ct_;
            accept(ct_);
        }
        
        return true;
    }

    bool CommandLineArgumentsParser::parseCDDBID()
    {
        if(ct_.empty())
            throw ParseException("Expected CDDB disc ID");
        else
        {
            parsed_.disc_id = (unsigned int)strtoll(ct_.c_str(), (char**)NULL, 16);
            accept(ct_);
        }
        
        return true;
    }

    bool CommandLineArgumentsParser::parseTemplate()
    {
        if(ct_.empty())
            throw ParseException("Expected template string");
        else
        {
            parsed_.templ = ct_;
            accept(ct_);
        }
        
        return true;
    }

    bool CommandLineArgumentsParser::parseCommand()
    {
        if(ct_ == TAGLOOKUP_ID)
        {
            parsed_.command = TAGLOOKUP_COMMAND_ID;
            accept(TAGLOOKUP_ID);
            parseCDDBCategory();
            parseCDDBID();
            parseFiles();
            return true;
        }
        
        if(ct_ == TAGLOOKUP_SEQUENCE)
        {
            parsed_.command = TAGLOOKUP_COMMAND_SEQUENCE;
            accept(TAGLOOKUP_SEQUENCE);
            parseFiles();
            return true;
        } 
        
        if(ct_ == TAGLOOKUP_RENAME)
        {
            parsed_.command = TAGLOOKUP_COMMAND_RENAME;
            accept(TAGLOOKUP_RENAME);
            parseTemplate();
            parseFiles();
            return true;
        }
        
        if(ct_ == TAGLOOKUP_LIST)
        {
            parsed_.command = TAGLOOKUP_COMMAND_LIST;
            accept(TAGLOOKUP_LIST);
            parseFiles();
            return true;
        }

        if(ct_ == TAGLOOKUP_HELP || ct_ == TAGLOOKUP_HELP_SHORT)
        {
            parsed_.command = TAGLOOKUP_COMMAND_HELP;
            accept(ct_);
            return true;
        }
        
        if(ct_.empty())
            throw ParseException("Expected command");
            
        throw ParseException("Unknown command");
    }
    
    bool CommandLineArgumentsParser::parseFiles()
    {
        // TODO: there should be a nicer way of doing this
        const std::vector<std::string>::const_iterator e = argv_.end();

        std::vector<std::string> &fs = parsed_.files;
        
        for(std::vector<std::string>::const_iterator i = ca_;
            i != e;
            ++i)
        {
            fs.push_back(*i);
            accept(*i);
        }
        
        if(fs.empty())
            throw ParseException("Expected at least one file");
            
        return true;
    }
    
    struct CopyItems
        : public std::unary_function<char * &, void>
    {
        std::vector<std::string> string_list;
        
        void operator() (char * & s)
        {
            string_list.push_back(std::string(s));
        }
    };
    
    CommandLineArgumentsParser::CommandLineArgumentsParser(int argc, char* argv[])
    {
        CopyItems ci = for_each(argv, argv + argc, CopyItems());
        
        argv_ = ci.string_list;
        
        ca_ = argv_.begin();
        ct_ = argv_.front();
        
        parseArguments();
    }
    
    const CommandLineArguments & CommandLineArgumentsParser::getArguments()
    {
        return parsed_;
    }
}

