/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * command_line.h
 * Describes a parser for command line arguments and the returned parsed 
 * structure
 *
 * BNF of the command line arguments:
 *  CommandLine ::= flags command
 * 
 * flags ::= flag
 *        |  flag flags
 * 
 * flag ::= '-c' cache_mode
 *       |  '-s' server
 *       |  '-p' port
 *       |  '-P' protocol
 * 
 * command ::= 'id' category disc_id files
 *          |  'sequence' files
 *          |  'rename' template
 *          |  'list' files
 * 
 * cache_mode ::= 'on' | 'off' | 'only'
 * 
 * server ::= string
 * 
 * port ::= integer
 * 
 * protocol ::= 'cddbp' | 'http' | 'proxy'
 * 
 * category ::= string
 * 
 * disc_id ::= hex
 * 
 * hex ::= string
 *
 * files ::= file
 *       |  file files

 * template ::= string
 */

#ifndef TAGLOOKUP_COMMAND_LINE_H
#define TAGLOOKUP_COMMAND_LINE_H

#include <string>
#include <vector>
#include <exception>

#include "cddb_query.h"

namespace TagLookup
{
    enum Command
    {
        TAGLOOKUP_COMMAND_NONE,
        TAGLOOKUP_COMMAND_ID,
        TAGLOOKUP_COMMAND_SEQUENCE,
        TAGLOOKUP_COMMAND_RENAME,
        TAGLOOKUP_COMMAND_LIST,
        TAGLOOKUP_COMMAND_HELP
    };

    enum Protocol
    {
        TAGLOOKUP_PROTOCOL_NONE,
        TAGLOOKUP_PROTOCOL_CDDBP,
        TAGLOOKUP_PROTOCOL_HTTP,
        TAGLOOKUP_PROTOCOL_HTTP_PROXY
    };

    struct CommandLineArguments
    {
        Command command;
        bool cache_mode_specified;
        Cache cache_mode;
        bool server_specified;
        std::string server;
        bool port_specified;
        unsigned int port;
        bool protocol_specified;
        Protocol protocol;
        
        std::string category;
        unsigned int disc_id;
        std::vector<std::string> files;
        std::string templ; // rename template
        
        CommandLineArguments()
            : command(TAGLOOKUP_COMMAND_NONE), 
              cache_mode_specified(false), 
              cache_mode(ON), 
              server_specified(false), 
              server(), 
              port_specified(false), 
              port(0), 
              protocol_specified(false), 
              protocol(TAGLOOKUP_PROTOCOL_NONE), 
              category(), 
              disc_id(0), 
              files(), 
              templ()
        {
        }
    };

    class CommandLineArgumentsParser
    {
    public:
        CommandLineArgumentsParser()
        {
        }
        
        CommandLineArgumentsParser(int argc, char* argv[]);
        const CommandLineArguments & getArguments();
    private:
        CommandLineArguments parsed_;
        
        std::vector<std::string> argv_;
        std::vector<std::string>::const_iterator ca_; // current argument in argv
        std::string ct_; // current token

        void accept(const std::string &str);
        void acceptEOF();
        bool parseArguments();
        bool parseFlags();
        bool parseFlag();
        bool parseCacheMode();
        bool parseServer();
        bool parsePort();
        bool parseProtocol();
        bool parseCDDBCategory();
        bool parseCDDBID();
        bool parseTemplate();
        bool parseCommand();
        bool parseFiles();
    };
    
    class ParseException
        : public std::exception
    {
    public:
        ParseException(const std::string &msg)
        {
            msg_ = "Parse Exception: ";
            msg_.append(msg);
        }

        ~ParseException() throw()
        {
        }
        
        const std::string &what()
        {
            return msg_;
        }
        
    private:
        std::string msg_;
    };
}
#endif

