/*
 * Copyright (c) Alex Allan <alex X kamaz.org.uk> 
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * tag.h
 * Provides the prototype to a number of classes related to file tags
 */

#ifndef TAGLOOKUP_TAG_H
#define TAGLOOKUP_TAG_H

#include <vector>
#include <utility>
#include <string>
#include <exception>

#include <ostream>

#include <taglib/fileref.h>

#include "tracks.h"
#include "cddb_query.h"

namespace TagLookup
{
    const std::string TOKEN = "%";
    const std::string ARTIST_TOKEN = "%a";
    const std::string GENRE_TOKEN = "%g";
    const std::string SONG_TITLE_TOKEN = "%s";    
    const std::string ALBUM_TITLE_TOKEN = "%t";    
    const std::string YEAR_TOKEN = "%y";    
    const std::string TRACK_NUMBER = "%n";    

    struct Offsets
    {
        OffsetList offset_list;
        unsigned int length;
    };

    class TagException
        : public std::exception
    {
    public:
        TagException(const std::string & msg)
            : msg_(msg)
        {}
        
        ~TagException() throw()
        {
        }
        
        virtual const char * what() const throw()
        {
            return msg_.c_str();
        }
    
    private:
        std::string msg_;
    }; // class TagException
    
    enum ProblemType
    {TAGLOOKUP_NONE = 0, TAGLOOKUP_READONLY = 1, TAGLOOKUP_INVALID_FILE = 2, TAGLOOKUP_READONLY_AND_INVALID_FILE = 3};
    
    /**
     * Describes a problem file. Returned by the 'checkFiles' method.
     */
    struct ProblemFile
    {
        std::string path;
        ProblemType type;
    };
    
    /**
     * Shows how file names will be changed when the renameFiles method is called.
     */
    struct OldNameNewName
    {
        std::string old_name;
        std::string new_name;
    };
    
    typedef std::vector<ProblemFile> ProblemFiles;
    
    class FileDisc
    {
        /**
         * FileDisc class
         * Represents a disc make up of files (with tags).
         */
    
    public:
        FileDisc(const std::vector<std::string> & file_path_list);

        static const std::string ProblemTypeToString(const ProblemType & problem_type);

        virtual ~FileDisc() throw ()
        {
        }
        
        // Actions:
        const Offsets & generateOffsets(int inital_offset);
        
        /**
         * Tag the files in the list by the fit of the their lengths.
         */
        void tagByBestFit(const Disc& disc);

        /**
         * Tag the files in the order that they appear in the FileDisc. If this
         * is longer than cddb_disc or visa versa, then no exception is raised.
         */
        void tagByOrder(const Disc& disc);

        /**
         * Prints the details of a file disc to an out stream.
         */
        void printDetails(std::ostream &os);

        /**
         * Rename the files contained within FileDisc using the given pattern,
         * where:
         * - %a is Artist
         * - %g is Genre
         * - %s is Song Title
         * - %t is Album title
         * - %y is Year; and
         * - %n is Track number
         */
        const std::vector<OldNameNewName> & generateNewNames(const std::string& templ);
        void renameFiles(const std::vector<OldNameNewName> &new_names);
        
        /**
         * Checks that files are valid for update. Returns a list of problems
         * if there are any that can't be tagged. ProblemFiles.empty() will be
         * true if there are none.
         */ 
        const ProblemFiles & checkFiles();

    private:
        struct PathFileRef
        {
            std::string path;
            TagLib::FileRef file;
        };

        std::vector<PathFileRef> files_;
        std::vector<OldNameNewName> new_names_;
        
        Offsets of_;
        ProblemFiles problems_;
        
        void stripTag(const TagLib::FileRef &fr);
        void setTag(const Disc &disc, const Track &track, TagLib::FileRef &fr);

        struct InstantiateTagLibFiles;
        struct OffsetsCounter;
        struct DiscCopyWithLengths;
        struct DiffFromX;
        struct PrepareNewNames;
        struct RenameFiles;
        struct PrintFileDisc;
        struct CheckValidity;
        
    }; // class FileDisc
    
} // namespace TagLookup

#endif

